			----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    16:08:21 09/12/2015 
-- Design Name: 
-- Module Name:    mcu_spi_flash - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
--use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity mcu_spi_flash is
    Port ( 	flash_io_wr : in  STD_LOGIC;
				flash_io_rd : in  STD_LOGIC;
				flash_cs_wr : in  STD_LOGIC;
				data_in : in  STD_LOGIC_VECTOR (7 downto 0);
				data_out : out  STD_LOGIC_VECTOR (7 downto 0);
				clk : in  STD_LOGIC;
				FLASH_CK : out  STD_LOGIC;
				FLASH_CS : out  STD_LOGIC;
				FLASH_SI : out  STD_LOGIC;
				FLASH_SO : in  STD_LOGIC
		);
end mcu_spi_flash;

architecture Behavioral of mcu_spi_flash is
	signal shift		: std_logic_vector(7 downto 0) := (others => '0');
	signal shift_in	: std_logic_vector(7 downto 0) := (others => '0');
	signal flash_data	: std_logic_vector(7 downto 0) := (others => '0');
	signal spi_ck_en	: std_logic := '0';
	signal spi_nce		: std_logic := '1';
	signal spi_state	: std_logic_vector( 2 downto 0) := (others => '0');
	signal next_spi_state	: std_logic_vector( 2 downto 0) := (others => '0');
	signal counter		: std_logic_vector(2 downto 0) := (others => '0');
	signal flash_init : std_logic;
	signal flash_done : std_logic;
	signal reg_out		: std_logic_vector(7 downto 0) := (others => '0');
begin
	FLASH_CK <= clk and spi_ck_en;				-- gated FLASH clock
	FLASH_CS <= spi_nce;									-- active low FLASH chip select/chip enable
	FLASH_SI <= shift(7);									-- MSB output to spi
	flash_init <='1' when (flash_io_wr='1' or flash_io_rd='1') else '0';
	data_out <= reg_out;
	
	-- advance state machine from state to state
	run_sm : process (clk, flash_init)
	begin
		if (flash_init = '0') then
			spi_state <= "000";								-- Initial state
		elsif rising_edge(clk) then
			spi_state <= next_spi_state;					-- next state
		end if;
	end process;

	process (clk,flash_cs_wr)
	begin
		if rising_edge(clk) then
			if (flash_cs_wr='1') then
				spi_nce <= data_in(0);
			end if;
		end if;
	end process;
	
	process (clk,flash_io_wr,flash_io_rd)
	begin
		if falling_edge(clk) then
			case spi_state is
				when"000"=>
					spi_ck_en <= '0';										-- Stop clock to FLASH
					--spi_nce <= '1';										-- Deselect FLASH		end if;
					flash_done <= '1';									-- FLASH comms done	end process;
					if flash_init = '1' then
						next_spi_state <= "001";						-- select next state
					end if;
				when"001"=>
					reg_out <= shift_in;
					shift <= data_in;										
					flash_done <= '0';									-- FLASH comms not done
					spi_ck_en <= '1';										-- enable FLASH clock
					--spi_nce <= '0';										-- Select FLASH
					counter <= "000";										-- reset counter
					next_spi_state <= "010";							-- select next state
				when"010" =>
					counter <= counter + 1;								-- count to next bit
					shift <= shift(6 downto 0) & '1';				-- shift other bits left
					shift_in <= shift_in(6 downto 0) & FLASH_SO;
					if (counter="111") then
						next_spi_state <= "011";
						spi_ck_en <= '0';										-- Stop clock to FLASH
					end if;
				when"011" =>
					--spi_nce <= '1';										-- Deselect FLASH		end if;
					flash_done <= '1';									-- FLASH comms done	end process;
					--shift_in <= shift_in(6 downto 0) & FLASH_SO;
					next_spi_state <= "100";
				when"100" =>
					if flash_init = '0' then
						next_spi_state <= "000";						-- select next state
					end if;
				when others =>												-- default
					next_spi_state <= "000";
				end case;
			end if;
		end process;
					
end Behavioral;

