-------------------------------------------------------------------------------
--	PSG AY-3-8910 for MSX
--
--
--
--
-------------------------------------------------------------------------------

library ieee;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity psg is
	port(
		clk	: in std_logic;
		reset	: in std_logic;
		cpu_din	:in  std_logic_vector(7 downto 0);
		cpu_dout	:out  std_logic_vector(7 downto 0);
		
		port0_wr : in std_logic;
		port1_wr : in std_logic;
		port2_rd : in std_logic;
		
		dac_out	:out  std_logic_vector(7 downto 0)
	);
end psg;
architecture RTL of psg is

	signal	clk_count	: std_logic_vector(7 downto 0);
	signal	docalc	: std_logic;
	signal	noise_seed	: std_logic_vector(19 downto 0);
	signal	noise_temp	: std_logic_vector(19 downto 0);
	signal	noise	: std_logic;
	signal	pstate	: std_logic_vector(2 downto 0);
	signal	mask_noise	: std_logic;
	signal	mask_tone	: std_logic;
	signal 	volume_in : std_logic_vector(3 downto 0);
	signal 	freq_in : std_logic_vector(11 downto 0);
	signal 	counter_in : std_logic_vector(11 downto 0);
	signal 	dacadd : std_logic_vector(7 downto 0);
	signal 	dacval : std_logic_vector(7 downto 0);
	signal	dacreg : std_logic_vector(7 downto 0);
	signal counter_a : std_logic_vector(11 downto 0);
	signal counter_b : std_logic_vector(11 downto 0);
	signal counter_c : std_logic_vector(11 downto 0);
	signal counter_n : std_logic_vector(11 downto 0);
	signal	phase_a : std_logic;
	signal	phase_b : std_logic;
	signal	phase_c : std_logic;
	signal	phase : std_logic;
	signal	phase_out : std_logic;
	signal counter_out : std_logic_vector(11 downto 0);
	
	signal reg_freq_a : std_logic_vector(11 downto 0);
	signal reg_freq_b : std_logic_vector(11 downto 0);
	signal reg_freq_c : std_logic_vector(11 downto 0);
	signal reg_freq_n : std_logic_vector(4 downto 0);
	signal reg_volume_a : std_logic_vector(4 downto 0);
	signal reg_volume_b : std_logic_vector(4 downto 0);
	signal reg_volume_c : std_logic_vector(4 downto 0);
	signal reg_mask : std_logic_vector(7 downto 0);
	signal reg_envcycle : std_logic_vector(15 downto 0);
	signal reg_env : std_logic_vector(3 downto 0);
	
	signal addr_reg : std_logic_vector(3 downto 0);
	signal cpu_access_state : std_logic_vector(2 downto 0);
	signal psg_out_val : std_logic_vector(7 downto 0);

begin
	docalc <= '1' when (clk_count="00000000") else '0';
	dac_out <= dacreg;
	cpu_dout <= psg_out_val;
	
	process (clk,reset,clk_count)
	begin
		if rising_edge(clk) then
		if (reset='1') then
			clk_count <= (others => '0');
		else
			if ( clk_count = "10001110" ) then	-- 32000000/1.7897725MHz/16 -1
			--if ( clk_count = "00001000" ) then	-- for sim
				clk_count <= (others => '0');
			else
				clk_count <= clk_count+'1';
			end if;
		end if;
		end if;
	end process;
	
	process (clk,addr_reg,reset,cpu_din,cpu_access_state,port0_wr,port1_wr)
	begin
		if rising_edge(clk) then
		if (reset='1') then
			cpu_access_state <= (others => '0');
			addr_reg <= (others => '0');
			reg_freq_a <= "000011111110";	-- 0x0fe (O4A)
			--reg_freq_a <= "000000001111";	-- 0x00f (O8A#)
			reg_freq_b <= (others => '0');
			reg_freq_c <= (others => '0');
			reg_freq_n <= (others => '0');
			reg_volume_a <="01111";
			reg_volume_b <= (others => '0');
			reg_volume_c <= (others => '0');
			reg_mask <= "00111110";
			reg_envcycle <= (others => '0');
			reg_env <= (others => '0');
		else
			case cpu_access_state is
			when "000" =>
				if (port0_wr='1') then
					addr_reg <= cpu_din(3 downto 0);
					cpu_access_state <= "001";
				end if;
				if (port1_wr='1') then
					case addr_reg is
						when "0000"=>
							reg_freq_a(7 downto 0) <= cpu_din;
						when "0001"=>
							reg_freq_a(11 downto 8) <= cpu_din(3 downto 0);
						when "0010"=>
							reg_freq_b(7 downto 0) <= cpu_din;
						when "0011"=>
							reg_freq_b(11 downto 8) <= cpu_din(3 downto 0);
						when "0100"=>
							reg_freq_c(7 downto 0) <= cpu_din;
						when "0101"=>
							reg_freq_c(11 downto 8) <= cpu_din(3 downto 0);
						when "0110" =>
							reg_freq_n <= cpu_din(4 downto 0);
						when "0111" =>
							reg_mask <= cpu_din;
						when "1000" =>
							reg_volume_a <= cpu_din(4 downto 0);
						when "1001" =>
							reg_volume_b <= cpu_din(4 downto 0);
						when "1010" =>
							reg_volume_c <= cpu_din(4 downto 0);
						when "1011" =>
							reg_envcycle(7 downto 0) <= cpu_din;
						when "1100" =>
							reg_envcycle(15 downto 8) <= cpu_din;
						when "1101" =>
							reg_env <=  cpu_din(3 downto 0);
						when others =>
					end case;
					cpu_access_state <= "001";
				end if;
				
			when "001" =>
				if ( (port0_wr='0') and (port1_wr='0') ) then
					cpu_access_state <= "000";
				end if;
			when others=>
			end case;
		end if;
		end if;
	end process;

	process (addr_reg,reg_freq_a,reg_freq_b,reg_freq_c,
		reg_mask,reg_volume_a,reg_volume_b,reg_volume_c,reg_envcycle,
		reg_env)
	begin
		case addr_reg is
			when "0000" =>
				psg_out_val <= reg_freq_a(7 downto 0);
			when "0001" =>
				psg_out_val <= "0000" & reg_freq_a(11 downto 8);
			when "0010" =>
				psg_out_val <= reg_freq_b(7 downto 0);
			when "0011" =>
				psg_out_val <= "0000" & reg_freq_b(11 downto 8);
			when "0100" =>
				psg_out_val <= reg_freq_c(7 downto 0);
			when "0101" =>
				psg_out_val <= "0000" & reg_freq_c(11 downto 8);
			when "0111" =>
				psg_out_val <=reg_mask;
			when "1000" =>
				psg_out_val <="000"&reg_volume_a;
			when "1001" =>
				psg_out_val <="000"&reg_volume_b;
			when "1010" =>
				psg_out_val <="000"&reg_volume_c;
			when "1011" =>
				psg_out_val <=reg_envcycle(7 downto 0);
			when "1100" =>
				psg_out_val <=reg_envcycle(15 downto 8);
			when "1101" =>
				psg_out_val <="0000" & reg_env;
			when others=>
				psg_out_val <=(others => 'X');
			end case;
	end process;

	process (clk,reset,clk_count,pstate,dacval,dacadd,counter_out,noise_temp)
	begin
		if ( rising_edge(clk)  ) then
		if (reset='1') then
			noise_seed <= "00001111111111111111";
			counter_a <= (others => '0');
			counter_b <= (others => '0');
			counter_c <= (others => '0');
			counter_n <= (others => '0');

			pstate <= (others => '0');
			dacreg <= (others => '0');
			dacval <=  (others => '0');
			phase_a <= '0';
			phase_b <= '0';
			phase_c <= '0';
		else 
			case pstate is
				when "000" =>
					counter_a <= counter_out;
					dacval <= dacval+dacadd;
					phase_a <= phase_out;
					pstate <= pstate +'1';
				when "001" =>
					counter_b <= counter_out;
					dacval <= dacval+dacadd;
					phase_b <= phase_out;
					pstate <= pstate +'1';
				when "010" =>
					counter_c <= counter_out;
					dacval <= dacval+dacadd;
					phase_c <= phase_out;
					pstate <= pstate +'1';
				when "011" =>
					if (counter_out="000000000000") then
						noise_seed  <= '0' & noise_temp(19 downto 1);
					end if;
					counter_n <= counter_out;
					pstate <= pstate +'1';
				when "100" =>
					dacreg <= dacval;
					pstate <= pstate +'1';
				when "101" =>
					dacval <=  (others => '0');
					if (docalc ='1') then
						pstate <= (others => '0'); 
					end if;
				when others =>
					pstate <= (others => '0');
			end case;
			 
		end if;
		end if;
	end process;

	process(noise_seed,pstate,mask_tone,mask_noise,volume_in,phase,counter_in,freq_in,
		counter_a,phase_a,phase_b,phase_c,reg_freq_a,reg_freq_b,reg_freq_c,reg_mask,
		counter_b,counter_c,reg_volume_a,reg_volume_b,reg_volume_c,phase_out)
	begin
		case pstate is
			when "000" =>
				counter_in <= counter_a;
				phase <= phase_a;
				freq_in <= reg_freq_a;
				volume_in <= reg_volume_a(3 downto 0);
				mask_tone <= reg_mask(0);
				mask_noise <= reg_mask(3);
			when "001" =>
				phase <= phase_b;
				counter_in <= counter_b;
				freq_in <= reg_freq_b;
				volume_in <= reg_volume_b(3 downto 0);
				mask_tone <= reg_mask(1);
				mask_noise <= reg_mask(4);
			when "010" =>
				phase <= phase_c;
				counter_in <= counter_c;
				freq_in <= reg_freq_c;
				volume_in <= reg_volume_c(3 downto 0);
				mask_tone <= reg_mask(2);
				mask_noise <= reg_mask(5);
			when "011" =>
				phase <= 'X';
				counter_in <= counter_n;
				freq_in <= "0000000" & reg_freq_n;
				volume_in <= (others => 'X');
				mask_tone <= 'X';
				mask_noise <= 'X';

			when others =>
				phase <= 'X';
				freq_in <= (others => 'X');
				volume_in <= (others => 'X');
				counter_in <= (others => 'X');
				mask_noise <= 'X';
				mask_tone <= 'X';
		end case;
		
		noise_temp <= noise_seed(19) & noise_seed(18) &
				(noise_seed(17) xor noise_seed(0)) &
				noise_seed(16 ) & noise_seed ( 15) &
				(noise_seed(14) xor noise_seed(0) ) &
				noise_seed(13 downto 0);
		noise <= noise_seed(0);
		
		if (counter_in /= freq_in ) then
			counter_out <= counter_in +'1';
			phase_out <= phase;
		else
			counter_out <= (others => '0');
			phase_out <= not phase;
		end if;

		if ( ((phase_out='1') and (mask_tone='0')) or ((noise='1') and (mask_noise='0')) ) then
			dacadd <= "00" & volume_in & "00";
		else
			dacadd <=  (others => '0');
		end if;
	end process;
	

	
end rtl;

