----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    09:44:12 09/20/2015 
-- Design Name: 
-- Module Name:    system - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
--use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity system is
    Port ( 	clk : in  STD_LOGIC;
				reset : in STD_LOGIC;
				mcu_mreq : in  STD_LOGIC;
           mcu_iorq : in  STD_LOGIC;
           mcu_rd : in  STD_LOGIC;
           mcu_wr : in  STD_LOGIC;
           mcu_intack : in  STD_LOGIC;
           mcu_mr : in  STD_LOGIC;
           mcu_addr : in  STD_LOGIC_VECTOR (15 downto 0);
           mcu_din : in  STD_LOGIC_VECTOR (7 downto 0);
           mcu_wait : out  STD_LOGIC;
           mcu_reset : out  STD_LOGIC;
			  MCU_clk_posedge : in  STD_LOGIC;
           sram_we : out  STD_LOGIC;
			  sram_sel : out  STD_LOGIC;
			  sram_addr : out  STD_LOGIC_VECTOR (10 downto 0);
           uart_enrx : out  STD_LOGIC;
           uart_entx : out  STD_LOGIC;
           uart_stbyrx : in  STD_LOGIC;
           uart_stbytx : in  STD_LOGIC;
			  uart_sel : out  STD_LOGIC;
           spi_cs : in  STD_LOGIC;
           spi_mosi : in  STD_LOGIC;
           spi_miso : out  STD_LOGIC;
           spi_clk : in  STD_LOGIC;

			  DC_addr : out STD_LOGIC_VECTOR (20 downto 0);
			  DC_dout : out STD_LOGIC_VECTOR (31 downto 0);
			  --DC_din : in STD_LOGIC_VECTOR (31 downto 0);
			  DC_valid  : in STD_LOGIC;
			  DC_cmdready : in STD_LOGIC;
			  DC_we : out STD_LOGIC;
			  DC_cmd : out STD_LOGIC;
			  DC_mask : out STD_LOGIC_VECTOR (3 downto 0);
			  DC_sel : out STD_LOGIC;

				PORT0_RD : out  STD_LOGIC;
				PORT1_RD : out  STD_LOGIC;
				PORT0_WR : out  STD_LOGIC;
				PORT1_WR : out  STD_LOGIC;
				vdp_ena  : out  STD_LOGIC;
				
				sys_sel : out  STD_LOGIC;
				sys_dout : out  STD_LOGIC_VECTOR (7 downto 0);
				test_out : out  STD_LOGIC_VECTOR (15 downto 0);
				
				flash_io_wr : out  STD_LOGIC;
				flash_io_rd : out  STD_LOGIC;
				flash_cs_wr : out  STD_LOGIC;

				psg_port0 : out  STD_LOGIC;
				psg_port1 : out  STD_LOGIC;
				psg_port2 : out  STD_LOGIC;

				led_out : out  STD_LOGIC;
			  lcd_sel : out  STD_LOGIC;
           lcd_data : out  STD_LOGIC_VECTOR (19 downto 0));

end system;

architecture Behavioral of system is
	signal busval  : STD_LOGIC_VECTOR(10 downto 0);
	signal uart_status : STD_LOGIC_VECTOR(7 downto 0);
	signal tmpreg : STD_LOGIC_VECTOR(7 downto 0) := (others => '0'); 
	signal mcu_reset_count : STD_LOGIC_VECTOR(4 downto 0) := (others => '0'); 
	signal breakpoint : STD_LOGIC_VECTOR(15 downto 0) := (others => '0');
	SIGNAL breakpoint_count : STD_LOGIC_VECTOR(3 downto 0) := (others => '0');

		-- ===== SPI
	signal spi_reg : STD_LOGIC_VECTOR(23 downto 0) := (others => '0');
	signal spi_count : STD_LOGIC_VECTOR(4 downto 0) := (others => '0'); 
	signal spi_mosi_i : STD_LOGIC;
	signal spi_count_save : STD_LOGIC_VECTOR(7 downto 0) := (others => '0'); 
	signal lcd_data_reg : STD_LOGIC_VECTOR(15 downto 0) := (others => '0'); 
	signal test_reg : STD_LOGIC_VECTOR(15 downto 0) := (others => '0'); 
	signal spi_command : STD_LOGIC_VECTOR(7 downto 0) := (others => '0'); 
	--signal write_ready_flag : STD_LOGIC;	
	signal CLK_8M : STD_LOGIC;
	signal lcd_wr_b : STD_LOGIC_VECTOR(4 downto 0) := (others => '1'); 
	signal lcd_wr_1 : STD_LOGIC;
	signal lcd_rs : STD_LOGIC;
	signal lcd_reset : STD_LOGIC; 
	signal vdp_enables : STD_LOGIC:='0';
	signal b_internal : STD_LOGIC_VECTOR(15 downto 0);
	signal b_extnal : STD_LOGIC_VECTOR(15 downto 0);
	signal countup : STD_LOGIC;
	--signal countreset : STD_LOGIC;
	
	-- ===== RAM/DRAM CONTROL
	signal MCU_MODE :STD_LOGIC :='0';	--monitor or msx
	signal sram_cs_0 : STD_LOGIC;
	signal sram_wr_0 : STD_LOGIC;
	signal dram_cs_0 : STD_LOGIC;
	signal dram_addr_0 : STD_LOGIC_VECTOR(20 downto 0);
	signal dram_wr_0 : STD_LOGIC;
	
	signal dram_cs_1 : STD_LOGIC;
	signal dram_wr_1 : STD_LOGIC;

	signal sram_cs : STD_LOGIC;
	signal sram_wr : STD_LOGIC;
	signal dram_cs : STD_LOGIC;
	signal dram_wr : STD_LOGIC;
	signal addr_page : STD_LOGIC_VECTOR(1 downto 0);
	
	signal slotv  : STD_LOGIC_VECTOR(1 downto 0);

	signal dram_page0 : STD_LOGIC_VECTOR(7 downto 0) := (others => '0'); 
	signal dram_page1 : STD_LOGIC_VECTOR(7 downto 0);
	signal dram_din8 : STD_LOGIC_VECTOR(7 downto 0);
	
	signal dstate : STD_LOGIC_VECTOR(2 downto 0) := (others => '0');
	signal iwait : STD_LOGIC;
	signal dram_cmd : STD_LOGIC;
	signal isRam : STD_LOGIC;

	-- ===== SLOT
	signal msx_slot_page0 : STD_LOGIC_VECTOR(8 downto 0);
	signal msx_slot_page1 : STD_LOGIC_VECTOR(8 downto 0);
	signal msx_slot_page2 : STD_LOGIC_VECTOR(8 downto 0);
	signal msx_slot_page3 : STD_LOGIC_VECTOR(8 downto 0);

	
		-- ==== 8255
	SIGNAL PPI_REGA : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
	SIGNAL PPI_REGB : STD_LOGIC_VECTOR(7 downto 0);
	SIGNAL PPI_REGC : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');

begin
	
	busval <= mcu_wr & mcu_iorq & mcu_mreq & mcu_addr(7 downto 0);
	uart_enrx <= '1' when (busval="01000111111") else '0';	--3Fh I/O read
	uart_entx <= '1' when (busval="11000100001") else '0';	--21h I/O write
	uart_sel <= '1' when (busval="01000111111") else '0';	--3FH UART data
	uart_status <= "0000" & uart_stbytx &"00" & uart_stbyrx;
	
	lcd_data <= '0' & lcd_reset & lcd_wr_1 & lcd_rs & lcd_data_reg;
	
	PORT0_RD <= '1' when (busval="01010011000") else '0';	--98h I/O read
	PORT0_WR <= '1' when (busval="11010011000") else '0';	--98h I/O write
	PORT1_RD <= '1' when (busval="01010011001") else '0';	--99h I/O read
	PORT1_WR <= '1' when (busval="11010011001") else '0';	--99h I/O write
	
	flash_io_rd <= '1' when (busval="01000101000") else '0';	--28h I/O read
	flash_io_wr <= '1' when (busval="11000101000") else '0';	--28h I/O write
	flash_cs_wr <= '1' when (busval="11000101001") else '0';	--29h I/O write
	
	psg_port0 <= '1' when ( (mcu_wr='1') and (mcu_iorq='1') and (mcu_addr(7 downto 0)="10100000")) else '0';
	psg_port1 <= '1' when ( (mcu_wr='1') and (mcu_iorq='1') and (mcu_addr(7 downto 0)="10100001")) else '0';
	--psg_port2 <= '1' when ( (mcu_wr='0') and (mcu_iorq='1') and (mcu_addr(7 downto 0)="10100010")) else '0';
	psg_port2 <= '0';


	-- MEMORY
	sram_cs <= '0' when (MCU_MODE = '1') else sram_cs_0;
	sram_addr <= MCU_Addr(10 downto 0);
	sram_wr <= '0' when (MCU_MODE = '1') else sram_wr_0;
	dram_cs <= dram_cs_1 when (MCU_MODE = '1') else dram_cs_0;
	DC_addr(20 downto 13) <= dram_page1 when (MCU_MODE = '1') else dram_page0;
	DC_addr(12 downto 0) <= MCU_Addr(13 downto 1);
	dram_wr <= dram_wr_1 when (MCU_MODE = '1') else dram_wr_0;

	addr_page <= mcu_addr(15 downto 14);
	sram_cs_0 <= '1' when (addr_page & mcu_mreq ="001") else '0';
	sram_wr_0 <= '1' when(addr_page & mcu_mreq & mcu_wr ="0011") else '0';
	dram_cs_0 <= '1' when (addr_page & mcu_mreq ="011") else '0';
	dram_wr_0 <= '1' when(addr_page & mcu_mreq & mcu_wr ="0111") else '0';
	
	dram_cs_1 <= mcu_mreq;
	dram_wr_1 <= '1' when((mcu_mreq='1') and (mcu_wr='1') and (isRam='1')) else '0';

	mcu_wait <= iwait;
	sram_sel <= sram_cs;
	sram_we <= sram_wr;
	DC_dout <= mcu_din & mcu_din & mcu_din & mcu_din;
	DC_we <= dram_wr;
	DC_cmd <= dram_cmd;
	DC_sel <= dram_cs;
	lcd_sel <='0';
	

	
	mcu_reset <= '0' when (mcu_reset_count="00000") else '1';

	-- DRAM control

	process (clk,dstate,DC_valid,DC_cmdready,dram_cs,mcu_wr,MCU_clk_posedge)
	begin
		if rising_edge(clk) then
		case dstate is
			when "000" =>
				if ( MCU_clk_posedge='1' ) then
					if ( (dram_cs = '1') and (mcu_wr = '0') ) then
						dstate <= "001";
					end if;
					if ( dram_wr ='1' ) then
						dstate <= "011";
					end if;
				end if;
			when "001" =>
				dstate <= "010";
			when "010" =>
				if (DC_valid = '1' ) then
					dstate <= "000";
				end if;
			when "011" =>
				dstate <= "100";
			when "100" =>
				if (DC_cmdready = '1' ) then
					dstate <= "000";
				end if;
			when others =>
				dstate <="000";
		end case;
		end if;
		case dstate is
			when "000" =>iwait <= '0';dram_cmd <= '0';
			when "001" =>iwait <= '1';dram_cmd <= '1';
			when "010" =>iwait <= '1';dram_cmd <= '0';
			when "011" =>iwait <= '1';dram_cmd <= '1';
			when "100" =>iwait <= '1';dram_cmd <= '0';
			when others =>iwait <= '0';dram_cmd <= '0';
		end case;
	end process;

	process (MCU_Addr)
	begin
		case MCU_Addr(1 downto 0) is
			when "00" => DC_mask <= "1000";
			when "01" => DC_mask <= "0100";
			when "10" => DC_mask <= "1000";
			when "11" => DC_mask <= "0100";
			when others => DC_mask <= "XXXX";
		end case;
	end process;
	
	-- ===== SLOT
	process (PPI_REGA)
	begin
		case PPI_REGA(1 downto 0) is
			when"00" => msx_slot_page0 <="000000000";	--BIOS0
			when others => msx_slot_page0 <="011111111";
		end case;
		case PPI_REGA(3 downto 2) is
			when"00" => msx_slot_page1 <="000000001";	--BIOS1
			when"01" => msx_slot_page1 <="000000010";	--CART0
			when others => msx_slot_page1 <="011111111";
		end case;
		case PPI_REGA(5 downto 4) is
			when"00" => msx_slot_page2 <="100000100";	--RAM0
			when"01" => msx_slot_page2 <="000000011";	--CART1
			when others => msx_slot_page2 <="011111111";
		end case;
		case PPI_REGA(7 downto 6) is
			when"00" => msx_slot_page3 <="100000101";	--RAM1
			when others => msx_slot_page3 <="011111111";
		end case;
	end process;

	process (addr_page,msx_slot_page0,msx_slot_page1,msx_slot_page2,msx_slot_page3)
	begin
		case addr_page is
			when "00" => dram_page1 <= msx_slot_page0(7 downto 0); isRam <= msx_slot_page0(8);
			when "01" => dram_page1 <= msx_slot_page1(7 downto 0); isRam <= msx_slot_page1(8);
			when "10" => dram_page1 <= msx_slot_page2(7 downto 0); isRam <= msx_slot_page2(8);
			when "11" => dram_page1 <= msx_slot_page3(7 downto 0); isRam <= msx_slot_page3(8);
			when others => dram_page1 <= "XXXXXXXX"; isRam <= 'X';
		end case;
	end process;

	-- I/O READ
	process (busval,uart_status,mcu_din)
	begin
		if (mcu_intack='1') then
			sys_dout <= "11111111";	-- fz80΍ MODE1łoXRST38HoKv
			sys_sel <= '1';
		else
			case busval is
				when "01000100000" =>
					sys_dout <= uart_status;	--20H UART status
					sys_sel <= '1';
				when "01010101000" =>
					sys_dout <= PPI_REGA;	--A8H PPI REGA(slot select)
					sys_sel <= '1';
				when "01010101001" =>
					case PPI_REGC(3 downto 0) is
						when "1000"=>
							sys_dout <= test_reg( 7 downto 0) or "00001000";
						when others => sys_dout <= "11111111";
					end case;
					sys_sel <= '1';
				when "01010100010" =>
					sys_dout <= "11111111";	--A2H 
					sys_sel <= '1';
				when "01010101010" =>
					sys_dout <= PPI_REGC;	--AAH PPI REGC
					sys_sel <= '1';
				when others => sys_dout <= "XXXXXXXX";sys_sel<='0';
			end case;
		end if;
	end process;


	-- I/O WRITE
	process (clk,busval,mcu_din)
	begin
		if rising_edge(clk) then
		if (reset='1') then
			mcu_reset_count <= "11111";
			MCU_MODE <= '0';
		else
			if (mcu_reset_count/="00000") then
				mcu_reset_count <= mcu_reset_count-1;
			end if;
		end if;
		if ((MCU_MODE = '1') AND (MCU_clk_posedge='1') and (breakpoint = mcu_addr) AND (mcu_wr='0')) then
			
			--if (breakpoint_count = "0001") then
				--MCU_MODE <= '0';	-- mode reg
				--mcu_reset_count <= "11111";	--CPU Reset
			--	breakpoint_count <= "0000";
			--else
			--	breakpoint_count <= breakpoint_count+1;
			--end if;
		end if;
		case busval is
			when "11000100010" =>	--22H modereg
				tmpreg <= mcu_din;
				MCU_MODE <= mcu_din(1);	-- mode reg
				if (mcu_din(2)='1') then
					--CPU Reset
					mcu_reset_count <= "11111";
				end if;
			when "11000100011" =>	--23H breakpoint(l)
				breakpoint(7 downto 0) <= mcu_din;
			when "11000100100" =>	--24H breakpoint(h)
				breakpoint(15 downto 8) <= mcu_din;
			when "11000100101" =>	--25H
				dram_page0 <= MCU_Din;
			when "11010101000" =>	--A8H
				PPI_REGA <= MCU_Din;
			when "11010101010" =>	--AAH
				PPI_REGC <= MCU_Din;
			when "11010101011" =>	--ABH
				if (MCU_Din(7)='0') then
					case MCU_Din(3 downto 1) is
						when"100" => PPI_REGC(4) <= MCU_Din(0);
						when"101" => PPI_REGC(5) <= MCU_Din(0);
						when"110" => PPI_REGC(6) <= MCU_Din(0);
						when"111" => PPI_REGC(7) <= MCU_Din(0);
						when others =>
					end case;
				end if;
			when others =>
		end case;
		end if;
	end process;

-- ============== SPI ==================
	SPI_MISO <='0' when SPI_CS='0' else 'Z';
	spi_mosi_i <= SPI_MOSI;

	countup <= '1' when ( (spi_count ="10111")) else '0';
	
	lcd_reset <= '0' when ( ( spi_command(7 downto 4) = "0001" ) and (spi_command(0) = '0' ) ) else '1'; 
	lcd_rs <= spi_command(0);
	lcd_wr_1 <= lcd_wr_b(1);
	vdp_ena <= vdp_enables;
	--TESTREG <= test_reg;
	
	process (SPI_CLK,SPI_MOSI,SPI_CS)
	begin
		if (SPI_CS = '1' ) then
			spi_count<="00000";
		else
		if rising_edge(SPI_CLK) then
			spi_reg <= spi_reg(22 downto 0) & spi_mosi_i;
			if (spi_count = "10111") then	--23
				lcd_data_reg <= spi_reg(14 downto 0) & spi_mosi_i;
				spi_command <= spi_reg(22 downto 15);
				if (spi_reg(22 downto 19) = "0000" ) then
					lcd_wr_b <= lcd_wr_b(3 downto 0) & '0';
				end if;
				if (spi_reg(22 downto 19) = "0010" ) then
					vdp_enables <= spi_reg(15);
				end if;
				if (spi_reg(22 downto 19) = "0011" ) then
					test_reg <= spi_reg(14 downto 0) & spi_mosi_i;
				end if;
				spi_count <= "00000";
			else
				spi_count <= spi_count+1;
				lcd_wr_b <= lcd_wr_b(3 downto 0) & '1';
			end if;
		end if;
		end if;
	end process;
	


	led_out <= tmpreg(0);
	--led_out <= mcu_intack;
	test_out <= "00000000" & PPI_REGA;
end Behavioral;

